#include "precompiled.h"

/*
Copyright 2001 Bruce Dawson, Cygnus Software
For more information, visit http://www.cygnus-software.com/papers/
or e-mail: comments@cygnus-software.com

This code may be redistributed as long as this notice is kept intact.
*/

#include <cmath>

#include "imagelibrary/colourspace.h"

// Pull in the necessary libraries.
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glaux.lib")
#pragma comment(lib, "glu32.lib")
#pragma warning(disable : 4244)	// Ignore double to float conversion warnings.
#pragma warning(disable : 4305)	// Ignore const double to float conversion warnings.

// Enable this for better debugging of drawing.
//#define	SINGLEBUFFERED
//#define	DOTEAPOT

static float s_windowWidth, s_windowHeight;

static bool s_paused = true;

static bool s_showInside = false;
static bool s_showYAxis = false;
static bool s_showUAxis = false;
static bool s_showVAxis = false;
static bool s_showIAxis = false;
static bool s_showQAxis = false;

static float s_xAngle = 45.0;
static float s_yAngle = -45.0;
static float s_zAngle = 10;

const float k_xAnimDelta = 0.7, k_yAnimDelta = 0.07;
const float k_arrowDelta = 1.5;

static const float s_vertices[][3] =
{
	 0,  1,  1,
	 1,  1,  1,
	 0,  0,  1,
	 1,  0,  1,
	 0,  1,  0,
	 1,  1,  0,
	 0,  0,  0,
	 1,  0,  0,
};

static const int s_faces[][4] =
{
	{ 0, 2, 3, 1 },	// Front?
	{ 0, 1, 5, 4 },
	{ 0, 4, 6, 2 },
	{ 7, 6, 4, 5 },	// Back?
	{ 7, 5, 1, 3 },
	{ 7, 3, 2, 6 },
};

void DoVertex(float x, float y, float z)
{
	glColor3f(x, y, z);
	glVertex3f(x, y, z);
}

void RenderCube(void)
{
	glMatrixMode(GL_MODELVIEW);

	// Reset our modelview matrix
	glLoadIdentity();

	// Finally we translate our cube back on the z-axis (into the screen)
	glTranslatef(0, 0, -4);

	// Then we rotate and scale our cube. I feel
	// obliged to mention that, like
	// most rotation code, this code gives the impression that
	// rotation has three degrees of freedom, when in fact it
	// only has two. That is the root cause of gimbal lock.
	glRotatef(s_xAngle, 1, 0, 0);
	glRotatef(s_yAngle, 0, 1, 0);
	glRotatef(s_zAngle, 0, 0, 1);

	// Magic scale factor to make the cube fit nicely in
	// the standard -1 to 1 range
	float gScale = 1.5;
	glScalef(gScale, gScale, gScale);

	// First we center our 1x1 cube around the origin. This lets us
	// use zero to one RGB values as vertex coordinates.
	glTranslatef(-.5, -.5, -.5);

	float thinWidth = s_windowWidth / 120;
	float thickWidth = s_windowWidth / 70;
	if (s_showInside)
	{
		glPolygonMode(GL_FRONT, GL_LINE);
		glPolygonMode(GL_BACK, GL_FILL);
		glDisable(GL_CULL_FACE);
		glLineWidth(thinWidth);
	}
	else
	{
		glCullFace(GL_BACK);
		glPolygonMode(GL_FRONT, GL_FILL);
		glEnable(GL_CULL_FACE);
	}

	for (int face = 0; face < NUMELEMENTS(s_faces); ++face)
	{
		glBegin(GL_QUADS);
		for (int vertex = 0; vertex < 4; ++vertex)
		{
			int vertNumber = s_faces[face][vertex];
			DoVertex(s_vertices[vertNumber][0], s_vertices[vertNumber][1], s_vertices[vertNumber][2]);
		}
		glEnd();
	}

	// Draw the YUV and YIQ axes.
	glLineWidth(thickWidth);
	glBegin(GL_LINES);
	if (s_showYAxis)
	{
		DoVertex(1, 1, 1);
		DoVertex(0, 0, 0);
	}
	if (s_showUAxis)
	{
		TypeRGB<float> rgb1(TypeYUV<float>(0.5, 0.4, 0).ToRGB());
		TypeRGB<float> rgb2(TypeYUV<float>(0.5, -.4, 0).ToRGB());
		DoVertex(rgb1.r, rgb1.g, rgb1.b);
		DoVertex(rgb2.r, rgb2.g, rgb2.b);
	}
	if (s_showVAxis)
	{
		TypeRGB<float> rgb1(TypeYUV<float>(0.5, 0, 0.4).ToRGB());
		TypeRGB<float> rgb2(TypeYUV<float>(0.5, 0, -.4).ToRGB());
		DoVertex(rgb1.r, rgb1.g, rgb1.b);
		DoVertex(rgb2.r, rgb2.g, rgb2.b);
	}
	// I should really add some code to rotate the UV axis
	// onto the IQ axis. Oh well - maybe later.
	if (s_showIAxis)
	{
		TypeRGB<float> rgb1(TypeYIQ<float>(0.5, 0.4, 0).ToRGB());
		TypeRGB<float> rgb2(TypeYIQ<float>(0.5, -.4, 0).ToRGB());
		DoVertex(rgb1.r, rgb1.g, rgb1.b);
		DoVertex(rgb2.r, rgb2.g, rgb2.b);
	}
	if (s_showQAxis)
	{
		TypeRGB<float> rgb1(TypeYIQ<float>(0.5, 0, 0.4).ToRGB());
		TypeRGB<float> rgb2(TypeYIQ<float>(0.5, 0, -.4).ToRGB());
		DoVertex(rgb1.r, rgb1.g, rgb1.b);
		DoVertex(rgb2.r, rgb2.g, rgb2.b);
	}
	glEnd();
}

void CALLBACK RenderScene()
{
	glClearColor(0, 0, 0, 1);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glColor3f(1, 0, 0);

#ifdef	DOTEAPOT
	glPushMatrix();
	glTranslatef(s_windowWidth / 2, s_windowHeight / 2, 0);
	auxWireTeapot(50);
	glPopMatrix();
#endif

	RenderCube();

#ifdef	SINGLEBUFFERED
	glFlush();
#else
	auxSwapBuffers();
#endif
}

void CALLBACK ChangeSize(GLsizei w, GLsizei h)
{
	// Avoid divide by zero.
	if (h == 0)
		h = 1;

	s_windowWidth = w;
	s_windowHeight = h;
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	//calculate the aspect ratio of the window
	//we'll use a perspective matrix to view our scene
	gluPerspective(45, (GLfloat)w/(GLfloat)h, 0.1, 100);

	glEnable(GL_DEPTH_TEST);
}

void CALLBACK IdleFunction()
{
	if (!s_paused)
	{
		s_xAngle += k_xAnimDelta;
		s_yAngle += k_yAnimDelta;

		RenderScene();
	}
	Sleep(10);	// Limit the frame rate
}

void CALLBACK PauseKeyProc()
{
	s_paused = !s_paused;
}

void ToggleFullScreen(HWND Window)
{
	DWORD ExStyles = WS_EX_LEFT | WS_EX_LTRREADING | WS_EX_RIGHTSCROLLBAR;
	DWORD Styles = WS_OVERLAPPED | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_VISIBLE;
	int Top, Left, Width, Height;
	HWND Above = 0;
	static int OldTop = 0;
	static int OldLeft = 0;
	static int OldWidth = 320;
	static int OldHeight = 240;
	DWORD CurrentExStyles = GetWindowLong(Window, GWL_EXSTYLE);
	if (CurrentExStyles & WS_EX_TOPMOST)
	{
		Top = OldTop;
		Left = OldLeft;
		Width = OldWidth;
		Height = OldHeight;
		Styles |= WS_MAXIMIZEBOX | WS_MINIMIZEBOX | WS_SYSMENU | WS_CAPTION | WS_SIZEBOX;
		Above = HWND_NOTOPMOST;
	}
	else
	{
		RECT	WindowRect;
		GetWindowRect(Window, &WindowRect);
		OldWidth = WindowRect.right - WindowRect.left;
		OldHeight = WindowRect.bottom - WindowRect.top;
		OldTop = WindowRect.top;
		OldLeft = WindowRect.left;
		ExStyles |= WS_EX_TOPMOST;
		Styles |= WS_POPUP;
		Width = GetSystemMetrics(SM_CXSCREEN);
		Height = GetSystemMetrics(SM_CYSCREEN);
		Top = Left = 0;
		Above = HWND_TOPMOST;
	}
	SetWindowLong(Window, GWL_STYLE, Styles);
	SetWindowLong(Window, GWL_EXSTYLE, ExStyles);
	SetWindowPos(Window, Above, Left, Top, Width, Height, 0);
}

void CALLBACK FKeyProc()
{
	HWND Window = auxGetHWND();
	ToggleFullScreen(Window);
}

void CALLBACK BKeyProc()
{
	s_showInside = !s_showInside;
}

void CALLBACK YKeyProc()
{
	s_showYAxis = !s_showYAxis;
}

void CALLBACK UKeyProc()
{
	s_showUAxis = !s_showUAxis;
}

void CALLBACK VKeyProc()
{
	s_showVAxis = !s_showVAxis;
}

void CALLBACK IKeyProc()
{
	s_showIAxis = !s_showIAxis;
}

void CALLBACK QKeyProc()
{
	s_showQAxis = !s_showQAxis;
}

void CALLBACK LeftKeyProc()
{
	s_yAngle -= k_arrowDelta;
}

void CALLBACK RightKeyProc()
{
	s_yAngle += k_arrowDelta;
}

void CALLBACK UpKeyProc()
{
	s_xAngle -= k_arrowDelta;
}

void CALLBACK DownKeyProc()
{
	s_xAngle += k_arrowDelta;
}

#ifdef	WIN32
// I use WinMain as an entry point to stop a console window from
// opening up.
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, 
				LPSTR lpCmdLine, int nCmdShow)
#else
int main(int argc, char* argv[])
#endif
{
#ifdef	SINGLEBUFFERED
	auxInitDisplayMode(AUX_SINGLE | AUX_RGBA | AUX_DEPTH );
#else
	auxInitDisplayMode(AUX_DOUBLE | AUX_RGBA | AUX_DEPTH );
#endif
	auxInitPosition(100, 100, 450, 450);
	auxInitWindow("Press 'F' to toggle full screen.");

	auxReshapeFunc(ChangeSize);
	auxIdleFunc(IdleFunction);

	auxKeyFunc(AUX_F, FKeyProc);
	auxKeyFunc(AUX_f, FKeyProc);

	auxKeyFunc(AUX_B, BKeyProc);
	auxKeyFunc(AUX_b, BKeyProc);

	auxKeyFunc(AUX_Y, YKeyProc);
	auxKeyFunc(AUX_y, YKeyProc);

	auxKeyFunc(AUX_U, UKeyProc);
	auxKeyFunc(AUX_u, UKeyProc);

	auxKeyFunc(AUX_V, VKeyProc);
	auxKeyFunc(AUX_v, VKeyProc);

	auxKeyFunc(AUX_I, IKeyProc);
	auxKeyFunc(AUX_i, IKeyProc);

	auxKeyFunc(AUX_Q, QKeyProc);
	auxKeyFunc(AUX_q, QKeyProc);

	auxKeyFunc(AUX_LEFT, LeftKeyProc);
	auxKeyFunc(AUX_RIGHT, RightKeyProc);
	auxKeyFunc(AUX_UP, UpKeyProc);
	auxKeyFunc(AUX_DOWN, DownKeyProc);

	auxKeyFunc(AUX_SPACE, PauseKeyProc);

#ifdef	GDCDEMO
	FKeyProc();
#endif
	auxMainLoop(RenderScene);

	return 0;
}
